"""
    To test dtk_pre_process and dtk_pre_process through EMODTask
"""
import json
import os
import sys

import pytest
from idmtools.assets import Asset
from idmtools.entities.experiment import Experiment

from idmtools.builders import SimulationBuilder
from idmtools.core.platform_factory import Platform
from idmtools.entities.templated_simulation import TemplatedSimulations
from idmtools_test.utils.itest_with_persistence import ITestWithPersistence

from emodpy.defaults import EMODSir
from emodpy.emod_task import EMODTask

current_directory = os.path.dirname(os.path.realpath(__file__))
BIN_PATH = os.path.join(current_directory, "..", "examples", "inputs", "bin")
INPUT_PATH = os.path.join(current_directory, "inputs", "process")

sim_duration = 10   # in years
num_seeds = 1

expname = os.path.split(sys.argv[0])[1]  # expname will be file name


def param_update(simulation, param, value):
    return simulation.set_parameter(param, value)


@pytest.mark.comps
class TestEmodPrePostProcess(ITestWithPersistence):
    def setUp(self) -> None:
        self.case_name = os.path.basename(__file__) + "--" + self._testMethodName
        self.platform = Platform('COMPS2')
        print(self.case_name)

    def test_emod_pre_process(self):
        task = EMODTask.from_default(default=EMODSir(), eradication_path=os.path.join(BIN_PATH, "Eradication.exe"))
        # set campaign to None to not sending any campaign to comps since we are going to override it later with
        # dtk-pre-process, This is import step in COMPS2
        task.campaign = None

        # need set this flag to add python_script_path to command argument by EMODTask
        task.use_embedded_python = True

        # Load dtk_pre_process.py to COMPS Assets/python folder
        dtk_pre_process_asset = Asset(os.path.join(INPUT_PATH, "dtk_pre_process.py"), relative_path='python')
        task.common_assets.add_asset(dtk_pre_process_asset)

        # Load campaign_template.json to simulation which used by dtk_pre_process.py
        campaign_template_asset = Asset(os.path.join(INPUT_PATH, "campaign_template.json"))
        task.transient_assets.add_asset(campaign_template_asset)

        # Create TemplatedSimulations with task
        ts = TemplatedSimulations(base_task=task)
        # Create SimulationBuilder
        builder = SimulationBuilder()
        # Add sweep parameter to builder
        builder.add_sweep_definition(EMODTask.set_parameter_partial("Run_Number"), range(num_seeds))
        # Add another sweep parameter to builder
        builder.add_sweep_definition(EMODTask.set_parameter_partial("Base_Infectivity"), [0.6, 1.0, 1.5, 2.0])
        # Add builder to simulations
        ts.add_builder(builder)

        # Create experiment from template
        experiment = Experiment.from_template(ts, name=self.case_name)

        self.platform.run_items(experiment)
        self.platform.wait_till_done(experiment)

        for sim in experiment.simulations:
            file = self.platform.get_files(sim, ["stdOut.txt"])
            stdout = file['stdOut.txt'].decode('utf-8')
            self.assertIn("daily inport pressures are : [", stdout)
            self.assertIn("durations are : [", stdout)

    def test_emod_post_process(self):
        task = EMODTask.from_default(default=EMODSir(), eradication_path=os.path.join(BIN_PATH, "Eradication.exe"))
        # need set this flag to add python_script_path to command argument by EMODTask
        task.use_embedded_python = True

        # Load dtk_post_process.py to COMPS Assets/python folder
        dtk_post_process_asset = Asset(os.path.join(INPUT_PATH, "dtk_post_process.py"), relative_path='python')
        task.common_assets.add_asset(dtk_post_process_asset)

        # Create experiment from template
        experiment = Experiment.from_task(task, name=self.case_name)

        self.platform.run_items(experiment)
        self.platform.wait_till_done(experiment)

        for sim in experiment.simulations:
            files = self.platform.get_files(sim, ["output/infection_rate.json"])
            data = json.loads(files['output/infection_rate.json'])['Data']
            self.assertEqual(90, len(data))
