import json
import os

import pytest
from idmtools.analysis.analyze_manager import AnalyzeManager
from idmtools.analysis.download_analyzer import DownloadAnalyzer
from idmtools.assets import Asset
from idmtools.builders import SimulationBuilder
from idmtools.core import ItemType
from idmtools.core.platform_factory import Platform
from idmtools.entities.experiment import Experiment
from idmtools.entities.simulation import Simulation
from idmtools_test.utils.itest_with_persistence import ITestWithPersistence
from emodpy.emod_task import EMODTask
from emodpy.generic.serialization import add_serialization_timesteps, load_serialized_population
from emodpy.defaults import EMODSir
from examples.config_update_parameters import del_folder
from tests.config_update_parameters import config_update_params

current_directory = os.path.dirname(os.path.realpath(__file__))
BIN_PATH = os.path.join(current_directory, "..", "examples", "inputs", "bin")
INPUT_PATH = os.path.join(current_directory, "..", "examples", "serialization", "inputs")


@pytest.mark.smoke
@pytest.mark.comps
class TestSerialization(ITestWithPersistence):

    def setUp(self) -> None:
        self.case_name = os.path.basename(__file__) + "--" + self._testMethodName
        print(self.case_name)
        self.platform = Platform('COMPS2')

    @pytest.mark.long
    def test_serialization(self):
        # Step1: create experiment and simulations with serialization file in output
        sim_duration = 2  # in years
        num_seeds = 1
        task1 = EMODTask.from_default(default=EMODSir(), eradication_path=os.path.join(BIN_PATH, "Eradication.exe"))

        # Update bunch of config parameters
        config_update_params(task1)
        timesteps = [sim_duration * 365]
        add_serialization_timesteps(task=task1, timesteps=timesteps,
                                    end_at_final=True, use_absolute_times=False)

        builder = SimulationBuilder()
        builder.add_sweep_definition(task1.set_parameter_partial("Run_Number"), range(num_seeds))

        experiment1 = Experiment.from_builder(builder, task1, name=self.case_name + " create serialization")

        self.platform.run_items(experiment1)
        self.platform.wait_till_done(experiment1)
        self.assertTrue(experiment1.succeeded)

        # download serialized file for later reload similation

        # Cleanup the output path
        output_path = 'output'
        del_folder(output_path)

        # We want to download all the dtk state files and the InsetChart.json
        filenames = []
        for serialization_timestep in timesteps:
            filenames.append("output/state-" + str(serialization_timestep).zfill(5) + ".dtk")
        filenames.append('output/InsetChart.json')

        analyzers = [DownloadAnalyzer(filenames=filenames)]
        # self.generate_experiment()
        am = AnalyzeManager(platform=self.platform, ids=[(experiment1.id, ItemType.EXPERIMENT)], analyzers=analyzers)
        am.analyze()

        # Step2: Create new experiment and sim with previous serialized file
        task2 = EMODTask.from_default(default=EMODSir(), eradication_path=os.path.join(BIN_PATH, 'Eradication.exe'))
        reload_sim = Simulation.from_task(task2)
        # Add the serialized population file to the experiment
        serialized_population_file = Asset(
            absolute_path=os.path.join('output', experiment1.simulations[0].id, 'state-00730.dtk'))
        task2.common_assets.add_asset(serialized_population_file)

        # Serialization parameters: Enable serialization and reload the population
        reload_sim.task.set_parameter("Simulation_Duration", sim_duration * 365)
        state_file = 'state-' + str(sim_duration * 365).zfill(5) + '.dtk'
        load_serialized_population(task2, population_path="Assets",
                                   population_filenames=[state_file])
        experiment2 = Experiment(name=self.case_name + " realod serialization")
        experiment2.tags = {'idmtools': 'reload serialization', 'exp2': 'tag2'}
        experiment2.simulations.append(reload_sim)
        self.platform.run_items(experiment2)
        self.platform.wait_till_done(experiment2)
        self.assertTrue(experiment2.succeeded)

        # validate InsetChart.json from experiment2 has timestep = sim_duration * 365
        for sim in experiment2.simulations:
            files = self.platform.get_files(sim, ["output/InsetChart.json"])
            time_steps = json.loads(files['output/InsetChart.json'])['Header']['Timesteps']
            self.assertEqual(sim_duration * 365, time_steps)
